/*
 * Copyright (c) 2002 Jonathan Chen.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
#if 0
static char sccsid[] = "@(#)compat.c	8.1 (+IMAP RPICS) 8/8/02";
#endif
#endif /* not lint */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "config.h"
#include "def.h"
#include "extern.h"

void err(int eval, const char* fmt, ...) {
    va_list ap;
    int code = errno;

    va_start(ap, fmt);
    fprintf(stderr, "%s: ", "mailx");
    if (fmt != NULL) {
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, ": ");
    }
    fprintf(stderr, "%s\n", strerror(code));
    exit(eval);
    va_end(ap);
}


void warn(const char* fmt, ...) {
    va_list ap;
    int code = errno;

    va_start(ap, fmt);
    fprintf(stderr, "%s: ", "mailx");
    if (fmt != NULL) {
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, ": ");
    }
    fprintf(stderr, "%s\n", strerror(code));
    va_end(ap);
}

void errx(int eval, const char* fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    fprintf(stderr, "%s: ", "mailx");
    if (fmt != NULL) {
	vfprintf(stderr, fmt, ap);
    }
    fprintf(stderr, "\n");
    exit(eval);
    va_end(ap);
}


void warnx(const char* fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    fprintf(stderr, "%s: ", "mailx");
    if (fmt != NULL) {
	vfprintf(stderr, fmt, ap);
    }
    fprintf(stderr, "\n");
    va_end(ap);
}

#ifndef HAVE_UNSETENV
extern char **environ;
void unsetenv(const char* name) {
    char **cur, **move;

    for (cur = environ; *cur; cur++) {
	if (strncmp(*cur, name, strlen(name)+1) == 0) {
	    /* Found variable; move subsequent variables over it */
	    for (move = cur; *move; move++)
	      *move = *(move + 1);
	    cur--;
	}
    }
}
#endif

#ifndef HAVE_STRCASESTR
char* strcasestr(const char *s, const char *find) {
    char c, sc;
    size_t len;
    
    if ((c = *find++) != 0) {
	c = tolower(c);
	len = strlen(find);
	do {
	    do {
		if (( sc = *s++) == 0)
		  return (NULL);
	    } while (tolower(sc) != c);
	} while (strncasecmp(s, find, len) != 0);
	s--;
    }
    return (char*)s;
}
#endif

#ifndef HAVE_STRSEP
char* strsep(char **stringp, const char* delim)
{
    char *s;
    const char *spanp;
    int c, sc;
    char *tok;

    if ((s = *stringp) == NULL)
	return (NULL);
    for (tok = s;;) {
	c = *s++;
	spanp = delim;
	do {
	    if ((sc = *spanp++) == c) {
		if (c == 0)
		    s = NULL;
		else
		    s[-1] = 0;
		*stringp = s;
		return (tok);
	    }
	} while (sc != 0);
    }
    /* NOTREACHED */
}
#endif

#ifndef HAVE_STRLCPY
size_t strlcpy(char *dst, const char *src, size_t siz) {
    char *d = dst;
    const char *s = src;
    size_t n = siz;

    if (n != 0 && --n != 0) {
	do {
	    if ((*d++ = *s++) == 0)
		break;
	} while (--n != 0);
    }
    if (n == 0) {
	if (siz != 0) *d = '\0';
	while (*s++);
    }
    return(s - src - 1);
}
#endif

#ifndef HAVE_STRLCAT
size_t strlcat(char *dst, const char *src, size_t siz) {
    size_t l = strlen(dst);
    return strlcpy(dst+l, src, siz-l);
}
#endif

