/*
 * Copyright (c) 2002 Jonathan Chen.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
#if 0
static char sccsid[] = "@(#)imap.c	9.0 (+IMAP RPICS) 8/8/02";
#endif
#endif /* not lint */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pwd.h>
#include "config.h"
#include "imap.h"
#include "rcv.h"
#include "extern.h"
#ifdef HAVE_GSSAPI_GSSAPI_H
#include <gssapi/gssapi.h>
#endif
#ifdef HAVE_GSSAPI_GSSAPI_GENERIC_H
#include <gssapi/gssapi_generic.h>
#endif

#ifndef GSS_C_NT_HOSTBASED_SERVICE
#define GSS_C_NT_HOSTBASED_SERVICE gss_nt_service_name
#endif

#define VATYPE va_list*
#define TOVATYPE &

#ifndef HAVE_TIMEGM
extern time_t timezone;
#endif

#define DPRINTFx(A) printf A
#define DPRINTF(A)

struct fetch_intdata;

enum imapd_states {
    state_discon,
    state_conn,
    state_authen
};

static int imap_untagged(char* buf);
static int imapd_waitok(int tag);
static int imapd_do_gssauth(char* user, char* host);
static int imapd_docmd(int tag, enum imapd_states req_state, char* cmd, ...);
static int imapd_send(enum imapd_states req_state, char* cmd, ...);
static int imapd_vsend(char* cmd, va_list ap);
static char* imapd_recv(void);
static char* imapd_recvlen(int);
static int imapd_dofetch(char* prefix, int msgnum, va_list ap);
static char* imapd_readfield(char** buf);
static char* imapd_findattr(char* large, char* what);
static struct fetch_intdata* findfid(enum imap_fetchwhat what);
static char* cr2crlf(char*);
static char* crlf2cr(char*);
static char* itoa(int x);

static int getlines1(char* data);
static int getattr_uid(char* data, void* out, VATYPE ap);
static int getattr_flags(char* data, void* out, VATYPE ap);
static int getattr_rfc822_size(char* data, void* out, VATYPE ap);
static int getattr_envelope_from(char* data, void* out, VATYPE ap);
static int getattr_fromline_addr(char* data, void* out, VATYPE ap);
static int getattr_fromline_date(char* data, void* out, VATYPE ap);
static int getattr_header(char* data, void* out, VATYPE ap);
static int getattr_lines(char* data, void* out, VATYPE ap);
static int getattr_hasmime(char* data, void* out, VATYPE ap);
static int getattr_body(char* data, void* out, VATYPE ap);
static int getattr_envdate_str(char* data, void* out, VATYPE ap);
static int getattr_bodypeek_size(char* data, void* out, VATYPE ap);

#include "base64.c"
#define GSSAUTH_P_NONE      1
#define GSSAUTH_P_INTEGRITY 2
#define GSSAUTH_P_PRIVACY   4

static char* imap_header_ditem(VATYPE ap, char*, int);
static char* imap_bodypeek_siz_ditem(VATYPE ap, char*, int);
struct fetch_intdata {
    enum imap_fetchwhat what;
    char* ditem;
    char* (*difunc)(VATYPE ap, char* data, int reqres);
    int (*getattr)(char* buf, void* output, VATYPE ap);
};
static struct fetch_intdata _fid[] = {
    {IMAP_UID, "UID", NULL, getattr_uid},
    {IMAP_FLAGS, "FLAGS", NULL, getattr_flags},
    {IMAP_RFC822_SIZE, "RFC822.SIZE", NULL, getattr_rfc822_size},
    {IMAP_FROMLINE_ADDR, "ENVELOPE", NULL, getattr_fromline_addr},
    {IMAP_FROMLINE_DATE, "INTERNALDATE", NULL, getattr_fromline_date},
    {IMAP_ENVELOPE_FROM, "ENVELOPE", NULL, getattr_envelope_from},
    {IMAP_HEADER, NULL, imap_header_ditem, getattr_header},
    {IMAP_LINES, "BODYSTRUCTURE", NULL, getattr_lines},
    {IMAP_HASMIME, "BODYSTRUCTURE", NULL, getattr_hasmime},
    {IMAP_BODY,	"BODY[]", NULL, getattr_body}, 
    {IMAP_ENVDATE_STR, "ENVELOPE", NULL, getattr_envdate_str},
    {IMAP_BODYPEEK_SIZE, NULL, imap_bodypeek_siz_ditem, getattr_bodypeek_size},
    {IMAP_END, NULL, NULL, NULL}
};
#define DITM(F,N) (((F)->ditem)?((N<2)?((F)->ditem):NULL):((F)->difunc(&ap,data,N)))

static enum imapd_states imap_state = state_discon;
static int imap_sock;
static FILE *imaprd;
static int imap_cmd = 1;
static int imapint_cmd = 1288;
static int imap_svrnummsg = -1;
static char* currbox = NULL;
static char* imap_pw = NULL;


static int getattr_uid(char* data, void* out, VATYPE ap) {
    char* ret;
    int n;

    n = strtol(data, &ret, 10);
    if (n <= 0) n = 0;
    *(int*)out = n;
    if (*ret != '\0') return -1;
    return 0;
}

static int getattr_flags(char* data, void* out, VATYPE ap) {
    short fl = 0;
    if (strstr(data, "\\Deleted")) fl |= MDELETED;
    if (strstr(data, "\\Recent")) fl |= MNEW;
    if (strstr(data, "\\Seen")) fl |= MREAD;
    *(short*)out = (*(short*)out & ~(MDELETED|MNEW)) | fl;
    return 0;
}

static int getattr_rfc822_size(char* data, void* out, VATYPE ap) {
    char* ret;
    long n;

    n = strtol(data, &ret, 10);
    if (n <= 0) n = 0;
    *(long*)out = n;
    if (*ret != '\0') return -1;
    return 0;
}

static int getattr_envelope_from(char* data, void* out, VATYPE ap) {
    char* from;
    char *name, *mbox, *host;
    char* env;

    (void)imapd_readfield(&data);
    (void)imapd_readfield(&data);
    from = imapd_readfield(&data);
    from = imapd_readfield(&from);
    name = imapd_readfield(&from);
    (void) imapd_readfield(&from); /* what do we do with the route address? */
    mbox = imapd_readfield(&from);
    host = imapd_readfield(&from);
    if (name && mbox && host) {
	env = malloc(strlen(name) + strlen(mbox) + strlen(host) + 5);
	strcpy(env, name);
	strcat(env, " <");
	strcat(env, mbox);
	strcat(env, "@");
	strcat(env, host);
	strcat(env, ">");
	*(char**)out = env;
	return 0;
    } else if (mbox && host) {
	env = malloc(strlen(mbox) + strlen(host) + 2);
	strcpy(env, mbox);
	strcat(env, "@");
	strcat(env, host);
	*(char**)out = env;
	return 0;
    } else if (mbox) {
	env = malloc(strlen(mbox) + 1);
	strcpy(env, mbox);
	*(char**)out = env;
	return 0;
    } else {
	*(char**)out = NULL;
	return -1;
    }
}

static int getattr_fromline_date(char* data, void* out, VATYPE ap) {
    struct tm tm, *tm2;
    time_t t;
    char buf[1024];
    char* zo;
    int zoff = 0;

    zo = strrchr(data, ' ');
    *zo = '\0';
    zo++;
    zoff = atoi(zo);
    zoff = ((zoff%100) + ((zoff/100)*60)) * 60;
    if (strptime(data, " %d-%b-%Y %H:%M:%S", &tm) == NULL) {
	abort();
	return -1;
    }
#ifdef HAVE_TIMEGM
    t = timegm(&tm) - zoff;
#else
    t = mktime(&tm) - timezone - zoff;
#endif
    tm2 = localtime(&t);
    strftime(buf, 1023, "%a %b %e %H:%M:%S %Y", tm2);
    *(char**)out = strdup(buf);
    return 0;
}

static int getattr_fromline_addr(char* data, void* out, VATYPE ap) {
    char* from;
    char *mbox, *host;
    char* env;

    (void)imapd_readfield(&data);
    (void)imapd_readfield(&data);
    from = imapd_readfield(&data);
    from = imapd_readfield(&from);
    (void) imapd_readfield(&from);
    (void) imapd_readfield(&from); /* what do we do with the route address? */
    mbox = imapd_readfield(&from);
    host = imapd_readfield(&from);
    if (mbox && host) {
	env = malloc(strlen(mbox) + strlen(host) + 2);
	strcpy(env, mbox);
	strcat(env, "@");
	strcat(env, host);
	*(char**)out = env;
	return 0;
    } else if (mbox) {
	env = malloc(strlen(mbox) + 1);
	strcpy(env, mbox);
	*(char**)out = env;
	return 0;
    } else {
	*(char**)out = NULL;
	return -1;
    }
}

static int getattr_header(char* data, void* out, VATYPE ap) {
    char* hdr;
    char* start, *end;
    
    (void)va_arg(*ap, char*);
    hdr = va_arg(*ap, char*);
    start = data;
    while (1) {
	start = strcasestr(start, hdr);
	if (start == NULL) {
	    *(char**)out = '\0';
	    return -1;
	} else if ((start == data || start[-1] == '\n') && start[strlen(hdr)] == ':') {
	    start += strlen(hdr)+1;
	    while (*start == ' ') start++;
	    end = strchr(start, '\r');
	    if (end) *end = '\0';
	    *(char**)out = strdup(start);
	    return 0;
	} else {
	    start++;
	}
    }
}

static int getlines1(char* data) {
    char* ptype;
    char* stype;
    char* sz;
    int l;

    ptype = imapd_readfield(&data);
    stype = imapd_readfield(&data);

    if (ptype == NULL) {
	return 0;
    } else if (strcmp(ptype, "TEXT") == 0) {
	imapd_readfield(&data);
	imapd_readfield(&data);
	imapd_readfield(&data);
	imapd_readfield(&data);
	imapd_readfield(&data);
	sz = imapd_readfield(&data);
    } else if (strcmp(ptype, "MESSAGE") == 0 && strcmp(stype, "RFC822") == 0) {
	imapd_readfield(&data);
	imapd_readfield(&data);
	imapd_readfield(&data);
	imapd_readfield(&data);
	imapd_readfield(&data);
	imapd_readfield(&data);
	imapd_readfield(&data);
	sz = imapd_readfield(&data);
    } else {
	return 0;
    }
    if (!sz) {
	fprintf(stderr, "getlines1: imap conversation error.\n");
	return 0;
    }
    l = strtol(sz, &ptype, 10);
    if (*ptype != '\0')
	fprintf(stderr, "getlines1: imap conversation error.\n");
    if (l < 0) return 0;
    return l;
}

static int getattr_lines(char* data, void* out, VATYPE ap) {
    long lines = 0;
    char* tbuf;

    if (data == NULL || *data == '\0') {
	*(long*)out = 0;
	return -1;
    }
    if (*data == '(') {
	/* multipart */
	while (data && *data) {
	    long nlines;
	    tbuf = imapd_readfield(&data);
	    getattr_lines(tbuf, &nlines, NULL);
	    lines += nlines;
	}
    } else {
	lines = getlines1(data);
    }
    if (lines <= 0) {
	*(long*)out = 0;
	return 1;
    }
    *(long*)out = lines;
    return 0;
}

static int getattr_hasmime(char* data, void* out, VATYPE ap) {
    if (*data == '(') {
	*(long*)out |= MMIME;
    } else {
	*(long*)out &= ~MMIME;
    }
    return 0;
}

static int getattr_body(char* data, void* out, VATYPE ap) {
    *(char**)out = crlf2cr(data);
    return 0;
}

static int getattr_envdate_str(char* data, void* out, VATYPE ap) {
    char* date;

    date = imapd_readfield(&data);
    if (date == NULL)
	*(char**)out = NULL;
    else
	*(char**)out = strdup(date);
    return 0;
}

static int getattr_bodypeek_size(char* data, void* out, VATYPE ap) {
    *(char**)out = crlf2cr(data);
    return 0;
}

void imap_close() {
    if (currbox) {
	free(currbox);
	currbox = NULL;
    }
    if (imap_state != state_discon)
	imapd_docmd(++imapint_cmd, state_discon, "LOGOUT");
    imap_state = state_discon;
    fclose(imaprd);
    close(imap_sock);
    imap_sock = -1;
}

int imap_open() {
    char* imap_username;
    char* host;
    struct hostent *hostent;
    struct servent *servent;
    struct sockaddr_in addr;
    int try_count;

    imap_state = state_discon;
    imap_sock = -1;

    imap_username = getenv("USER");
    if (!(imap_username && *imap_username))
	imap_username = getlogin();
    if (! (imap_username && *imap_username)) {
	struct passwd *passwd;
	passwd = getpwuid(getuid());
	if (passwd && passwd->pw_name && *passwd->pw_name)
	    imap_username = passwd->pw_name;
    }
    if (!(imap_username && *imap_username)) {
	fprintf(stderr, "imap_open: Could not determine username\n");
	return -1;
    }

    host = getenv("MAILHOST");
    if (!(host && *host))
	host = value("MAILHOST");
    if (!(host && *host))
	host = MAILHOST;
    
    try_count = 0;
    {
	hostent = gethostbyname(host);
	if (try_count++ >= 5 || ((!hostent) && (h_errno != TRY_AGAIN))) {
	    fprintf(stderr, "imap_open: Could not determine mail server's address");
	    return -1;
	}
    } while (!hostent);

    bzero((char *) &addr, sizeof(addr));
    addr.sin_family = AF_INET;
    servent = getservbyname("imap", "tcp");
    if (servent)
	addr.sin_port = servent->s_port;
    else
	addr.sin_port = 143;

    imap_sock = socket(PF_INET, SOCK_STREAM, 0);
    if (imap_sock < 0) {
	perror("imap_open: socket");
	return -1;
    }

    for (;*hostent->h_addr_list;hostent->h_addr_list++) {
	memcpy(&addr.sin_addr, *hostent->h_addr_list, hostent->h_length);
	if (!connect(imap_sock, (struct sockaddr*) &addr, sizeof(addr)))
	    break;
    }
    if (!*hostent->h_addr_list) {
	perror("imap_open: connect");
	return -1;
    }
    imaprd = fdopen(imap_sock, "r");
    imap_state = state_conn;

    if (imapd_waitok(0)) {
	fprintf(stderr, "imap_open: bad greeting\n");
	return -1;
    }

    if (imapd_do_gssauth(imap_username, host) > 0) {
	imap_state = state_authen;
    }

    try_count = 0;
    while (imap_state != state_authen && try_count < 3) {
	if (!imap_pw) {
	    char msg[300];
	    sprintf(msg, "Enter imap password for %.100s@%.100s: ", imap_username, host);
	    imap_pw = strdup((getpass(msg)));
	}
	if (imapd_docmd(++imapint_cmd, state_conn, "LOGIN \"%.100s\" \"%.100s\"", imap_username, imap_pw)) {
	    free(imap_pw);
	    imap_pw = NULL;
	    try_count++;
	    fprintf(stderr, "LOGIN %.100s failed\n", imap_username);
	    continue;
	}
	imap_state = state_authen;
	break;
    }
    if (imap_state != state_authen) return -1;

    if (currbox) {
	if (imapd_docmd(++imapint_cmd, state_authen, "SELECT %s", currbox)) {
	    fprintf(stderr, "Unable to select folder '%s'\n", currbox);
	    return -1;
	}
    }
    return 0;
}

int imap_select(char* mailbox) {
    if (currbox) {
	free(currbox);
	currbox = NULL;
    }
    if (imapd_docmd(++imap_cmd, state_authen, "SELECT %s", mailbox)) {
	fprintf(stderr, "Unable to select folder '%s'\n", mailbox);
	return -1;
    }
    if (currbox) free(currbox);
    currbox = strdup(mailbox);
    return 0;
}

int imap_deleteuid(int uid) {
    if (imapd_docmd(++imap_cmd, state_authen, "UID STORE %d +FLAGS (\\Deleted)", uid)) {
	fprintf(stderr, "Unable to delete message #%d\n", uid);
	return -1;
    }
    return 0;
}

int imap_setflaguid(int uid, long flag) {
    char data[1024];

    if (flag & MSTATUS) {
	data[0] = '\0';
	if (flag & MDELETED) strcat(data, "\\Deleted ");
	if (flag & MREAD) strcat(data, "\\Seen ");
	if (flag & MMARK) strcat(data, "\\Flagged ");
	if (data[0] != '\0') {
	    if (imapd_docmd(++imap_cmd, state_authen, "UID STORE %d +FLAGS (%s)", uid, data)) {
		fprintf(stderr, "Unable to set flags for message #%d\n", uid);
		return -1;
	    }
	}
	data[0] = '\0';
	if (!(flag & MDELETED)) strcat(data, "\\Deleted ");
	if (!(flag & MREAD)) strcat(data, "\\Seen ");
	if (!(flag & MMARK)) strcat(data, "\\Flagged ");
	if (data[0] != '\0') {
	    if (imapd_docmd(++imap_cmd, state_authen, "UID STORE %d -FLAGS (%s)", uid, data)) {
		fprintf(stderr, "Unable to set flags for message #%d\n", uid);
		return -1;
	    }
	}
    }
    return 0;
}

#ifdef BAD_MAIL_SERVER_UGLY_HACK_XXX
    int imap_really_expunge_all = 0;
#endif

int imap_expunge() {
#ifdef BAD_MAIL_SERVER_UGLY_HACK_XXX
    if (!imap_really_expunge_all)
        imapd_docmd(++imap_cmd, state_authen, "STORE 1 -FLAGS (\\Deleted)");
    imap_really_expunge_all = 0;
#endif
    if (imapd_docmd(++imap_cmd, state_authen, "EXPUNGE")) {
	fprintf(stderr, "Unable to expunge mailbox\n");
	return -1;
    }
    return 0;
}

int imap_append(char* buf) {
    char* nbuf = cr2crlf(buf);
    int oldnum = imap_svrnummsg;
    int imap_count;

    if (imapd_docmd(++imap_cmd, state_authen, "APPEND %s (\\Seen) {%d}", currbox, strlen(nbuf)) != 1) {
	fprintf(stderr, "Unable to append message\n");
	free(nbuf);
	return -1;
    }
    if (imapd_send(state_authen, "%s\r\n", nbuf)) {
	fprintf(stderr, "Unable to send appended message\n");
	free(nbuf);
	return -1;
    }
    free(nbuf);
    if (imapd_waitok(imap_cmd) != 0) {
	fprintf(stderr, "Error while appending message\n");
	free(nbuf);
	return -1;
    }

    for (imap_count = 0;
	 oldnum == imap_svrnummsg && imap_count < 5 && (sleep(1) || 1);
	 imap_count++) {
	imapd_docmd(++imap_cmd, state_authen, "NOOP");
    }
    
    if (oldnum == imap_svrnummsg) {
	fprintf(stderr, "imap_append: server did not notify us of new message\n");
	return -1;
    }

    if (imap_fetch(imap_svrnummsg, IMAP_UID, &imap_count, NULL)) {
	fprintf(stderr, "imap_append: unable to fetch new message uid\n");
	return -1;
    }

    return imap_count;
}

int imap_msgcount() {
    imapd_docmd(++imap_cmd, state_authen, "NOOP");
    return imap_svrnummsg;
}

int imap_fetch(int msgnum, ...) {
    va_list ap;
    int a;

    va_start(ap, msgnum);
    a = imapd_dofetch("", msgnum, ap);
    va_end(ap);
    return a;
}

int imap_fetchuid(int msgnum, ...) {
    va_list ap;
    int a;

    va_start(ap, msgnum);
    a = imapd_dofetch("UID ", msgnum, ap);
    va_end(ap);
    return a;
}

static char* imap_header_ditem(VATYPE ap, char* data, int reqres) {
    static char a[1024];
    char *x;
    if (reqres == 0)
	strcpy(a, "BODY.PEEK[HEADER.FIELDS (");
    else
	strcpy(a, "BODY[HEADER.FIELDS (");
    if (reqres == 2) {
	strlcat(a, "\"", 1024);
	strlcat(a, data, 1024);
	strlcat(a, "\"", 1024);
    } else {
        x= va_arg((*ap), char*);
	strlcat(a, x, 1024);
    }
    strlcat(a, ")]", 1024);
    return a;
}

static char* imap_bodypeek_siz_ditem(VATYPE ap, char* data, int reqres) {
    static char a[1024];
    if (reqres == 0) {
	strcpy(a, "BODY.PEEK[]<");
	strlcat(a, itoa(va_arg(*ap, int)), 1024);
	strlcat(a, ".", 1024);
	strlcat(a, itoa(va_arg(*ap, int)), 1024);
	strlcat(a, ">", 1024);
    } else if (reqres == 1) {
	strcpy(a, "BODY[]<");
	strlcat(a, itoa(va_arg(*ap, int)), 1024);
	strlcat(a, ".", 1024);
	strlcat(a, itoa(va_arg(*ap, int)), 1024);
	strlcat(a, ">", 1024);
    } else {
	strcpy(a, "BODY[]<");
	strlcat(a, itoa(*(*(int**)*ap-2)), 1024);
	strlcat(a, ">", 1024);
    }
    return a;
}

static struct fetch_intdata* findfid(enum imap_fetchwhat what) {
    struct fetch_intdata *cfid;

    for (cfid = _fid; cfid->what != IMAP_END; cfid++) {
	if (cfid->what == what)
	    return cfid;
    }
    return NULL;
}

static char* itoa(int x) {
    static char a[1024];
    snprintf(a, 1024, "%d", x);
    return a;
}

static int imapd_dofetch(char* prefix, int msgnum, va_list ap) {
    char fetch[1024];
    enum imap_fetchwhat what;
    struct fetch_intdata *fid;
    va_list sap;
    char* buf, *tbuf;
    char* ditm;
    int getting;
    
    va_copy(sap, ap);
    fetch[0] = '\0';

    while (IMAP_END != (what = va_arg(ap, enum imap_fetchwhat))) {
	char* data;

	data = va_arg(ap, void*);
	fid = findfid(what);
	if (fid == NULL) {
	    fprintf(stderr, "imapd_dofetch: invalid parameters\n");
	    abort();
	    return -1;
	}
	ditm = DITM(fid,0);
	buf = strstr(fetch, ditm);
	if (!(buf && (buf == fetch || buf[-1] == ' ') &&
	      buf[strlen(fid->ditem)] == ' ')) {
	    strlcat(fetch, ditm, 1024);
	    strlcat(fetch, " ", 1024);
	}
    }
    
    while (fetch[strlen(fetch)-1] == ' ') fetch[strlen(fetch)-1] = '\0';
    
    if (imapd_send(state_authen, "%d %sFETCH %d (%s)\r\n", ++imap_cmd, prefix, msgnum, fetch)) {
	fprintf(stderr, "imapd_dofetch: imapd_send failed\n");
	return -1;
    }
    getting = 0;
redo:
    if (NULL == (tbuf = imapd_recv())) {
	fprintf(stderr, "imapd_dofetch: imapd_recv failed\n");
	return -1;
    }
    if (getting || (*tbuf == '*' && strstr(tbuf, " FETCH "))) {
	char *start;
	if (getting) {
	    buf = realloc(buf, strlen(buf)+strlen(tbuf)+1);
	    strcat(buf, tbuf);
	} else {
	    buf = tbuf;
	}
	if ((start = strrchr(tbuf, '{'))) {
	    int len;
	    len = strtol(start+1, NULL, 10);
	    if (getting) free(tbuf);
	    tbuf = imapd_recvlen(len);
	    buf = realloc(buf, strlen(buf)+strlen(tbuf)+1);
	    strcat(buf, tbuf);
	    free(tbuf);
	    getting = 1;
	    goto redo;
	} else {
	    if (getting) free(tbuf);
	}
    } else if (*tbuf == '*') {
	if (!imap_untagged(tbuf+2)) {
	    fprintf(stderr, "imapd_dofetch: unexpected response: %s\n", tbuf);
	    goto redo;
	}
	goto redo;
    }
    imapd_waitok(imap_cmd);

    va_copy(ap, sap);
     
    while (IMAP_END != (what = va_arg(ap, enum imap_fetchwhat))) {
	char* data;
	char* ret;
	va_list ap2;

	va_copy(ap2, ap);
	data = va_arg(ap, void*);
	fid = findfid(what);
	ditm = DITM(fid,1);
	ret = imapd_findattr(buf, ditm);
	if (!ret) {
	    char* nditm;
	    nditm = DITM(fid,2);
	    if (nditm) ret = imapd_findattr(buf, nditm);
	}
	if (!ret) {
	    fprintf(stderr, "imapd_dofetch: absent attribute '%s'\n", ditm);
	    abort();
	    free(buf);
	    return -1;
	}
	if (fid->getattr(ret, data, &ap2) < 0) {
DPRINTF(("imapd_dofetch: getattr failed on '%s'\n", ditm));
	    free(buf);
	    free(ret);
	    return -1;
	}
	free(ret);
    }
    free(buf);
    return 0;
}

static int imapd_send(enum imapd_states req_state, char* cmd, ...) {
    va_list ap;
    int ret;

    if (imap_state < req_state)
	imap_open();
    if (imap_state < req_state)
	return -1;
    va_start(ap, cmd);
    ret = imapd_vsend(cmd, ap);
    va_end(ap);
    return ret;
}

static int imapd_vsend(char* cmd, va_list ap) {
    char buf[1024];
    char *lbuf;
    int mlen = 1024;

    vsnprintf(buf, sizeof(buf), cmd, ap);
    lbuf = buf;
    while (strlen(lbuf)+1 == mlen) {
	if (lbuf != buf) free(lbuf);
	mlen = mlen*2;
	lbuf = malloc(mlen);
	vsnprintf(lbuf, mlen, cmd, ap);
    }
DPRINTF(("XXX: send: %s\n", buf));
    if (write(imap_sock, lbuf, strlen(lbuf)) != (ssize_t)strlen(lbuf)) {
	if (lbuf != buf) free(lbuf);
	return -1;
    }
    if (lbuf != buf) free(lbuf);
    return 0;
}

static jmp_buf alrmstop;
static void sigalrm(int signo) {
    fprintf (stderr, "imapd_recv: receive timeout!\n");
    longjmp(alrmstop, 1);
}

static char* imapd_recv() {
    char* buf, *tbuf;
    int len = 0;

    if (imap_state == state_discon) return NULL;
    tbuf = buf = malloc(1024);

    if (setjmp(alrmstop))
	goto bad_recv;
    (void)signal(SIGALRM, sigalrm);
    alarm(15);

    while (1) {
	if (fgets(tbuf, 1024, imaprd) == NULL) {
	    alarm(0);
bad_recv:
	    free(buf);
	    imap_state = state_discon;
	    return NULL;
	}
	alarm(0);
	if (strlen(tbuf) == 1023 && tbuf[1022] != '\n') {
	    len += 1023;
	    buf = realloc(buf, len+1025);
	    tbuf = buf + len;
	} else {
DPRINTF(("XXX: recv: %s\n", buf));
	    return buf;
	}
    }
}

static char* imapd_recvlen(int len) {
    char* buf;

    buf = malloc(len+1);
    len = fread(buf, 1, len, imaprd);
    if (len <= 0) {
	free(buf);
	return NULL;
    }
    buf[len] = '\0';
DPRINTF(("XXX: recl: %s\n", buf));
    return buf;
}

static int imapd_docmd(int tag, enum imapd_states req_state, char* cmd, ...) {
    va_list ap;
    char *tmp;

    va_start(ap, cmd);
    tmp = strdup(itoa(tag));
    if (imapd_send(req_state, tmp))
	goto bad;
    if (imapd_send(req_state, " "))
	goto bad;
    if (imapd_vsend(cmd, ap))
	goto bad;
    if (imapd_send(req_state, "\r\n"))
	goto bad;
    free(tmp);
    va_end(ap);
    return imapd_waitok(tag);
    bad:
    free(tmp);
    va_end(ap);
    return -1;
}

static int imapd_waitok(int tag) {
    char* buf = NULL;
    char* atag;

    while (buf == NULL) {
	if (NULL == (buf = imapd_recv())) {
	    return -1;
	} else if (buf[0] == '*' && buf[1] == ' ') {
	    if (tag == 0) {
		free(buf);
		return 0;
	    }
	    while (buf[strlen(buf)-1] == '\r' || buf[strlen(buf)-1] == '\n')
		buf[strlen(buf)-1] = '\0';
	    imap_untagged(buf+2);
	    free(buf);
	    buf = NULL;
	} else if (buf[0] == '+' && buf[1] == ' ') {
	    free(buf);
	    return 1;
	}
    }

    atag = itoa(tag);
    if (strstr(buf, atag) == buf && buf[strlen(atag)] == ' ') {
	free(buf);
	return 0;
    } else {
	return -1;
    }
}

static int imap_untagged(char* buf) {
    /* process untagged */
    char* tok;
    char* tmp;
    int tn;

    tok = strchr(buf, ' ');
    if (strcmp(buf, "BYE") == 0) {
	imap_state = state_discon;
	return 1;
    } else if ((strcmp(tok+1, "EXISTS") == 0 ||
		strcmp(tok+1, "EXISTS\r\n") == 0) &&
	     ((tn = strtol(buf, &tmp, 10)) || 1) && tmp == tok) {
	imap_svrnummsg = tn;
	return 1;
    } else if ((strcmp(tok+1, "EXPUNGE") == 0 ||
		strcmp(tok+1, "EXPUNGE\r\n") == 0) &&
	     (tn = strtol(buf, &tmp, 10) || 1) && tmp == tok) {
	return 1;
    } else if ((strcmp(tok+1, "RECENT") == 0 ||
		strcmp(tok+1, "RECENT\r\n") == 0) &&
	     (tn = strtol(buf, &tmp, 10) || 1) && tmp == tok) {
	return 1;
    }
    return 0;
}

static char* cr2crlf(char* buf) {
    int size = strlen(buf) + (strlen(buf)/40);
    int nsize = 0;
    char* ret;
    char* nret;
    char* last = buf;
    char* cr;

    nret = ret = malloc(size);
    for (cr = strchr(last, '\n'); cr != NULL; cr = strchr(last, '\n')) {
	nsize += (cr-last);
	if (nsize >= size) {
	    int diff = (nret - ret);
	    size += 80;
	    ret = realloc(ret, size);
	    nret = ret + diff;
	}
	memcpy(nret, last, cr-last);
	nret += (cr-last);
	last = cr+1;
	*(nret++) = '\r';
	*(nret++) = '\n';
    }
    if (last) {
	nsize += strlen(last);
	if (nsize >= size) {
	    int diff = (nret - ret);
	    size += 80;
	    ret = realloc(ret, size);
	    nret = ret + diff;
	}
	strcpy(nret, last);
    }
    return ret;
}

static char* crlf2cr(char* buf) {
    char *nret;
    char *ret;
    char *last = buf;
    char *cr;

    nret = ret = malloc(strlen(buf)+1);
    bzero(ret, strlen(buf)+1);
    for (cr = strchr(last, '\r'); cr != NULL; cr = strchr(last, '\r')) {
	if (cr[1] != '\n') { last = cr+1; continue; }
	memcpy(nret, last, cr-last);
	nret += cr-last;
	*(nret++) = '\n';
	last = cr+1;
	if (*last) last++;
    }
    if (last != NULL)
	strcat(nret, last);
    return ret;
}

static char* imapd_readfield(char** buf) {
    char* start;
    char* tbuf;
    char *o, *c, *q, *b;
    int tmp;

    if (*buf == NULL) return NULL;
    /* returns start of field, null-terminates field, set buf to next */
    while (**buf == ' ') (*buf)++;
    start = *buf;

    switch (*start) {
      case '{':
	tmp = strtol(start+1, &tbuf, 10);
	if (*tbuf != '}' || tmp <= 0) {
	    fprintf(stderr, "imapd_readfield: imap conversation error\n");
	    return NULL;
	}
	start = tbuf+3;
	if (strlen(start) < tmp) {
	    fprintf(stderr, "imapd_readfield: short field\n");
	    tmp = strlen(start);
	}
	(*buf) = start + tmp;
	**buf = '\0';
	(*buf)++;
	return start;
      case '(':
	start++;
	tbuf = start;
	tmp = 1;
	while (tmp > 0) {
	    o = strchr(tbuf, '(');
	    c = strchr(tbuf, ')');
	    q = strchr(tbuf, '"');
	    if ((o<c || c==NULL) && (o<q || q==NULL) && o != NULL) {
		tmp++;
		tbuf = o+1;
	    } else if ((c<o || o==NULL) && (c<q || q==NULL) && c != NULL) {
		tmp--;
		tbuf = c+1;
	    } else if ((q<c || c==NULL) && (q<o || o==NULL) && q != NULL) {
		tbuf = strchr(q+1, '"') + 1;
	    } else {
		fprintf(stderr, "imapd_readfield: imap conversation error\n");
		return NULL;
	    }
	}
	tbuf[-1] = '\0';
	*buf = tbuf;
	return start;
      case '"':
	start++;
	tbuf = strchr(start, '"');
	if (tbuf == NULL) {
	    fprintf(stderr, "imapd_readfield: imap conversation error\n");
	    return NULL;
	}
	*tbuf = '\0';
	*buf = tbuf+1;
	return start;
      default:
	tbuf = start;
	do {
	    o = strchr(tbuf, ' ');
	    c = strchr(tbuf, ')');
	    q = strchr(tbuf, '"');
	    b = strchr(tbuf, '[');
	    if ((q<c || c==NULL) && (q<o || o==NULL) && (q<b || b==NULL) && q != NULL) {
		tbuf = strchr(tbuf, '"')+1;
		continue;
	    } else if ((b<c || c==NULL) && (b<o || o==NULL) && (b<q || q==NULL) && b != NULL) {
		tbuf = strchr(tbuf, ']')+1;
		continue;
	    }
	    break;
	} while (1);

	if ((o < c || c == NULL) && o != NULL) {
	    tbuf = o;
	    *tbuf = '\0';
	    *buf = tbuf+1;
	} else if (c != NULL) {
	    tbuf = c;
	    *tbuf = '\0';
	    *buf = NULL;
	} else {
	    tbuf = start + strlen(start);
	    *buf = NULL;
	}
	if (strcmp(start, "NIL") == 0)
	    return NULL;
	return start;
    }
}

static char* imapd_findattr(char* large, char* what) {
    char* n;
    char* name;
    char* data;

    n = large = strdup(large);
    large = strchr(large, '(');
    if (!large) {
	fprintf(stderr, "imapd_findattr: imap conversation error\n");
	free(n);
	return NULL;
    }
    large++;

    while (large && *large != '\0') {
	name = imapd_readfield(&large);
	data = imapd_readfield(&large);
	if (name && strcasecmp(name, what) == 0) {
	    if (data) {
		data = strdup(data);
		free(n);
		return data;
	    } else {
		fprintf(stderr, "imapd_findattr: imap conversation error\n");
		free(n);
		return NULL;
	    }
	}
    }
	
    free(n);
    return NULL;
}

int imapd_do_gssauth(char* user, char* host) {
    gss_buffer_desc request_buf, send_token;
    gss_buffer_t sec_token;
    gss_name_t target_name;
    gss_ctx_id_t context;
    gss_qop_t quality;
    int cflags;
    OM_uint32 maj_stat, min_stat;
    char buf1[8192], buf2[8192], server_conf_flags;
    unsigned long buf_size;
    char* buf;
    int result;

    /* first things first: get an imap ticket for host */
    sprintf(buf1, "imap@%s", host);
    request_buf.value = buf1;
    request_buf.length = strlen(buf1) + 1;
    maj_stat = gss_import_name(&min_stat, &request_buf, GSS_C_NT_HOSTBASED_SERVICE,
        &target_name);
    if (maj_stat != GSS_S_COMPLETE) {
	fprintf(stderr, "imapd_do_gssauth: Cannot get service name for %s\n", buf1);
	return -1;
    }

    if (imapd_send(state_conn, "%d AUTHENTICATE GSSAPI\r\n", ++imapint_cmd)) {
	perror("imapd_do_gssauth: imapd_send");
	return -1;
    }

    if (!(buf = imapd_recv())) {
	perror("imapd_do_gssauth: imapd_recv");
	return -1;
    } else if (*buf != '+') {
	fprintf(stderr, "imapd_do_gssauth: gssapi not supported\n");
	return -1;
    }

    /* now start the security context initialisation loop... */
    sec_token = GSS_C_NO_BUFFER;
    context = GSS_C_NO_CONTEXT;
    do {
        send_token.length = 0;
	send_token.value = NULL;
        maj_stat = gss_init_sec_context(&min_stat, 
				        GSS_C_NO_CREDENTIAL,
            				&context, 
					target_name, 
					GSS_C_NO_OID, 
					GSS_C_MUTUAL_FLAG | GSS_C_SEQUENCE_FLAG, 
					0, 
					GSS_C_NO_CHANNEL_BINDINGS, 
					sec_token, 
					NULL, 
					&send_token, 
					NULL, 
					NULL);
        if (maj_stat!=GSS_S_COMPLETE && maj_stat!=GSS_S_CONTINUE_NEEDED) {
	    fprintf(stderr, "imapd_do_gssauth: Error exchanging credentials\n");
            gss_release_name(&min_stat, &target_name);
            /* wake up server and await NO response */
	    if (imapd_send(state_conn, "\r\n")) return -1;
	    if ((buf = imapd_recv())) free(buf);
            return 0;
        }
        to64frombits((unsigned char*)buf1, send_token.value, send_token.length);
        gss_release_buffer(&min_stat, &send_token);

	if (imapd_send(state_conn, buf1) || imapd_send(state_conn, "\r\n")) {
	    perror("imapd_do_gssauth: imapd_send");
	    return -1;
	}

        if (maj_stat == GSS_S_CONTINUE_NEEDED) {
	    if (!(buf = imapd_recv())) {
		perror("imapd_do_gssauth: imapd_recv");
		return -1;
	    }
	    request_buf.length = from64tobits(buf2, buf + 2, sizeof(buf2));
	    free(buf);
	    if (request_buf.length == -1)	/* in case of bad data */
		request_buf.length = 0;
	    request_buf.value = buf2;
	    sec_token = &request_buf;
        }
    } while (maj_stat == GSS_S_CONTINUE_NEEDED);
    gss_release_name(&min_stat, &target_name);

    /* get security flags and buffer size */
    if (!(buf = imapd_recv())) {
	perror("imapd_do_gssauth: imapd_recv");
	return -1;
    }

    request_buf.length = from64tobits(buf2, buf + 2, sizeof(buf2));
    free(buf);
    if (request_buf.length == -1)	/* in case of bad data */
	request_buf.length = 0;
    request_buf.value = buf2;

    maj_stat = gss_unwrap(&min_stat, context, 
			  &request_buf, &send_token, &cflags, &quality);
    if (maj_stat != GSS_S_COMPLETE) {
        fprintf(stderr, "imapd_do_gssauth: Couldn't unwrap security level data\n");
        gss_release_buffer(&min_stat, &send_token);
        return -1;
    }
    /* first octet is security levels supported. We want none, for now */
    server_conf_flags = ((char *)send_token.value)[0];
    if ( !(((char *)send_token.value)[0] & GSSAUTH_P_NONE) ) {
        fprintf(stderr, "imapd_do_gssauth: Server requires integrity and/or privacy\n");
        gss_release_buffer(&min_stat, &send_token);
        return -1;
    }
    ((char *)send_token.value)[0] = 0;
    buf_size = ntohl(*((long *)send_token.value));
    /* we don't care about buffer size if we don't wrap data */
    gss_release_buffer(&min_stat, &send_token);

    /* now respond in kind (hack!!!) */
    buf_size = htonl(buf_size); /* do as they do... only matters if we do enc */
    memcpy(buf1, &buf_size, 4);
    buf1[0] = GSSAUTH_P_NONE;
    strcpy(buf1+4, user); /* server decides if princ is user */
    request_buf.length = 4 + strlen(user) + 1;
    request_buf.value = buf1;
    maj_stat = gss_wrap(&min_stat, context, 0, GSS_C_QOP_DEFAULT, &request_buf,
        &cflags, &send_token);
    if (maj_stat != GSS_S_COMPLETE) {
        fprintf(stderr, "imapd_do_gssauth: Error creating security level request\n");
        return -1;
    }
    to64frombits((unsigned char*)buf1, send_token.value, send_token.length);

    result = 1;
    if (imapd_send(state_conn, buf1) || imapd_send(state_conn, "\r\n")) {
	perror("imapd_do_gssauth: imapd_send");
	result = -1;
    } else {
	if (imapd_waitok(imapint_cmd)) {
	    result = -1;
	} 
    }

    /* flush security context */
    maj_stat = gss_delete_sec_context(&min_stat, &context, &send_token);
    if (maj_stat != GSS_S_COMPLETE) {
	fprintf(stderr, "imapd_do_gssauth: Error releasing credentials\n");
	return -1;
    }
    /* send_token may contain a notification to the server to flush
     * credentials. RFC 1731 doesn't specify what to do, and since this
     * support is only for authentication, we'll assume the server
     * knows enough to flush its own credentials */
    gss_release_buffer(&min_stat, &send_token);

    return result;
}
